﻿using InABox.Core;

namespace InABox.Database;

public delegate void LogEvent(LogType type, string message);

public interface IProviderFactory
{
    string URL { get; set; }
    Type[] Types { get; set; }

    void ForceRecreateViews();

    void Start();

    IProvider NewProvider(Logger logger);
}

public interface IProvider
{
    Logger Logger { get; set; }

    IEnumerable<object[]> List<T>(Filter<T>? filter = null, Columns<T>? columns = null, SortOrder<T>? sort = null, CoreRange? range = null) where T : Entity, new();
    
    CoreTable Query<T>(Filter<T>? filter = null, Columns<T>? columns = null, SortOrder<T>? sort = null, 
        CoreRange? range = null, bool log = false, bool distinct = false) where T : Entity, new();
    
    CoreTable Query(Type type, IFilter? filter = null, IColumns? columns = null, ISortOrder? sort = null, 
        CoreRange? range = null, bool log = false, bool distinct = false);
    
    /// <summary>
    /// Same as <see cref="Query{T}(Filter{T}, Columns{T}, SortOrder{T}, CoreRange, bool, bool)"/>, but only for deleted items
    /// </summary>
    /// <typeparam name="T"></typeparam>
    /// <param name="deletion"></param>
    /// <param name="filter"></param>
    /// <param name="columns"></param>
    /// <param name="sort"></param>
    /// <param name="range"></param>
    /// <param name="deleted"></param>
    /// <returns></returns>
    CoreTable QueryDeleted<T>(Deletion deletion, Filter<T>? filter = null, Columns<T>? columns = null, SortOrder<T>? sort = null, CoreRange? range = null, bool deleted = false) where T : Entity, new();

    T[] Load<T>(Filter<T>? filter = null, SortOrder<T>? sort = null, CoreRange? range = null) where T : Entity, new();

    void Save<T>(T entity) where T : Entity;
    void Save<T>(IEnumerable<T> entities) where T : Entity;

    void Save(Type type, Entity entity);
    void Save(Type type, IEnumerable<Entity> entities);

    void Delete<T>(T entity, string userID) where T : Entity, new();
    void Delete<T>(IEnumerable<T> entities, string userID) where T : Entity, new();

    void Purge<T>(T entity) where T : Entity;
    void Purge<T>(IEnumerable<T> entities) where T : Entity;

    void Purge(Deletion deletion);
    void Recover(Deletion deletion);

    List<Type> GetDeletionTypes(Deletion deletion);
}